package main

import (
	"github.com/gin-gonic/gin"
	log "github.com/sirupsen/logrus"
	"io/ioutil"
	"net/http"
	"random.chars.jp/git/image-board/store"
	"strconv"
)

type UserPayload struct {
	Username   string `json:"username"`
	ID         string `json:"id"`
	Privileged bool   `json:"privileged"`
}

type UserCreatePayload struct {
	Username   string `json:"username"`
	Password   string `json:"password"`
	Privileged bool   `json:"privileged"`
}

type UserUpdatePayload struct {
	Username string `json:"username"`
}

type TagUpdatePayload struct {
	Type string `json:"type"`
}

type ImageUpdatePayload struct {
	Source string `json:"source"`
}

var unauthorized = gin.H{"error": "not authorized"}
var denied = gin.H{"error": "permission denied"}

func registerAPI() {
	router.GET("/api", func(context *gin.Context) {
		context.JSON(http.StatusOK, store.Info{
			Revision:       instance.Revision,
			Compat:         instance.Compat,
			Register:       instance.Register,
			InitialUser:    instance.InitialUser,
			PermissionDir:  instance.PermissionDir,
			PermissionFile: instance.PermissionFile,
		})
	})

	router.GET("/api/single_user", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.SingleUser)
	})

	router.GET("/api/user", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.Users())
	})

	router.PUT("/api/user", func(context *gin.Context) {
		user, ok := getUser(context)
		if !instance.Register {
			if !ok {
				context.JSON(http.StatusForbidden, gin.H{
					"error": "user registration disallowed",
				})
				return
			}

			if !user.Privileged {
				context.JSON(http.StatusForbidden, denied)
				return
			}
		}

		var payload UserCreatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": err.Error(),
			})
			return
		}
		if !user.Privileged && payload.Privileged {
			context.JSON(http.StatusForbidden, denied)
			return
		}

		context.JSON(http.StatusOK, instance.UserAdd(payload.Username, payload.Privileged, payload.Password))
	})

	router.GET("/api/user/this", func(context *gin.Context) {
		info, ok := getUser(context)
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}
		context.JSON(http.StatusOK, UserPayload{
			Username:   info.Username,
			ID:         info.Snowflake,
			Privileged: info.Privileged,
		})
	})

	router.GET("/api/user/:flake", func(context *gin.Context) {
		info := instance.User(context.Param("flake"))
		context.JSON(http.StatusOK, UserPayload{
			Username:   info.Username,
			ID:         info.Snowflake,
			Privileged: info.Privileged,
		})
	})

	router.PATCH("/api/user/:flake", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, denied)
			return
		}

		var payload UserUpdatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": err.Error(),
			})
			return
		}

		instance.UserUsernameUpdate(flake, payload.Username)
	})

	router.DELETE("/api/user/:flake/password", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require privileged
		if !ok || !info.Privileged {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		flake := context.Param("flake")
		instance.UserPasswordUpdate(flake, "")
		instance.UserSecretRegen(flake)
	})

	router.PUT("/api/user/:flake/password", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, denied)
		}

		var newPass string
		if payload, err := context.GetRawData(); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": err.Error(),
			})
			return
		} else {
			newPass = string(payload)
		}

		if newPass == "" {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": "empty passwords are not allowed",
			})
			return
		}

		instance.UserPasswordUpdate(info.Snowflake, newPass)
		context.JSON(http.StatusOK, gin.H{
			"secret": instance.UserSecretRegen(info.Snowflake),
		})
	})

	router.GET("/api/username/:name", func(context *gin.Context) {
		info := instance.UserUsername(context.Param("name"))
		payload := UserPayload{
			Username:   info.Username,
			ID:         info.Snowflake,
			Privileged: info.Privileged,
		}
		context.JSON(http.StatusOK, payload)
	})

	router.POST("/api/username/:name/auth", func(context *gin.Context) {
		var password string

		if payload, err := context.GetRawData(); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": err.Error(),
			})
			return
		} else {
			password = string(payload)
		}

		username := context.Param("name")
		if instance.UserUsernamePasswordValidate(username, password) {
			context.String(http.StatusOK, instance.UserUsername(username).Secret)
		} else {
			context.JSON(http.StatusForbidden, denied)
		}
	})

	router.DELETE("/api/user/:flake", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		instance.UserDestroy(flake)
	})

	router.GET("/api/user/:flake/secret", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		context.String(http.StatusOK, instance.User(flake).Secret)
	})

	router.PUT("/api/user/:flake/secret", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		flake := context.Param("flake")
		if !info.Privileged && (info.Snowflake != flake) {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		context.String(http.StatusOK, instance.UserSecretRegen(flake))
	})

	router.GET("/api/user/:flake/image", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.UserImages(context.Param("flake")))
	})

	router.GET("/api/image/snowflake", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require privileged
		if !ok || !info.Privileged {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		context.JSON(http.StatusOK, instance.ImageSnowflakes())
	})

	router.GET("/api/image/snowflake/:flake", func(context *gin.Context) {
		flake := context.Param("flake")
		if _, err := strconv.Atoi(flake); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": "invalid snowflake",
			})
			return
		}
		context.JSON(http.StatusOK, instance.ImageSnowflake(flake))
	})

	router.GET("/api/image", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require privileged
		if !ok || !info.Privileged {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		context.JSON(http.StatusOK, instance.Images())
	})

	router.POST("/api/image", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		payload, err := context.FormFile("image")
		if err != nil {
			context.JSON(http.StatusInternalServerError, err)
			return
		}
		file, err := payload.Open()
		if err != nil {
			log.Errorf("Error while opening uploaded file %s, %s", payload.Filename, err)
			context.JSON(http.StatusInternalServerError, err)
			return
		}
		data, err := ioutil.ReadAll(file)
		if err != nil {
			log.Errorf("Error while reading uploaded file %s, %s", payload.Filename, err)
			context.JSON(http.StatusInternalServerError, err)
			return
		}
		image := instance.ImageAdd(data, info.Snowflake)
		if image.Hash == "" {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": "invalid image",
			})
			return
		}
		context.JSON(http.StatusOK, image)
	})

	router.GET("/api/image/:hash", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.Image(context.Param("hash")))
	})

	router.PATCH("/api/image/:hash", func(context *gin.Context) {
		user, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		info := instance.Image(context.Param("hash"))
		if !user.Privileged && (info.User != user.Snowflake) {
			context.JSON(http.StatusForbidden, denied)
			return
		}

		var payload ImageUpdatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": err.Error(),
			})
			return
		}

		instance.ImageSource(info.Hash, payload.Source)
	})

	router.DELETE("/api/image/:hash", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		hash := context.Param("hash")
		if !info.Privileged && (info.Snowflake != instance.Image(hash).User) {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		instance.ImageDestroy(hash)
	})

	router.GET("/api/image/:hash/file", func(context *gin.Context) {
		image, data := instance.ImageData(context.Param("hash"), false)
		if image.Hash == "" {
			context.JSON(http.StatusNotFound, gin.H{
				"error": "not found",
			})
			return
		}
		context.Data(http.StatusOK, "image/"+image.Type, data)
	})

	router.GET("/api/image/:hash/preview", func(context *gin.Context) {
		image, data := instance.ImageData(context.Param("hash"), true)
		if image.Hash == "" {
			context.JSON(http.StatusNotFound, gin.H{
				"error": "not found",
			})
			return
		}
		context.Data(http.StatusOK, "image/jpeg", data)
	})

	router.GET("/api/image/:hash/tag", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.ImageTags(context.Param("hash")))
	})

	router.PUT("/api/image/:hash/tag/:tag", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		hash := context.Param("hash")
		if !info.Privileged && (info.Snowflake != instance.Image(hash).User) {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		instance.ImageTagAdd(hash, context.Param("tag"))
	})

	router.DELETE("/api/image/:hash/tag/:tag", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		hash := context.Param("hash")
		if !info.Privileged && (info.Snowflake != instance.Image(hash).User) {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		instance.ImageTagRemove(hash, context.Param("tag"))
	})

	router.GET("/api/tag", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.Tags())
	})

	router.GET("/api/tag/:tag", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		context.JSON(http.StatusOK, instance.Tag(context.Param("tag")))
	})

	router.PUT("/api/tag/:tag", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		context.JSON(http.StatusOK, instance.TagCreate(context.Param("tag")))
	})

	router.DELETE("/api/tag/:tag", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, denied)
			return
		}
		instance.TagDestroy(context.Param("tag"))
	})

	router.GET("/api/tag/:tag/page", func(context *gin.Context) {
		tag := context.Param("tag")
		if !instance.MatchName(tag) {
			context.JSON(http.StatusBadRequest, denied)
			return
		}
		context.String(http.StatusOK, strconv.Itoa(instance.PageTotal("tag_"+tag)))
	})

	router.GET("/api/tag/:tag/page/:entry", func(context *gin.Context) {
		tag := context.Param("tag")
		if !instance.MatchName(tag) {
			context.JSON(http.StatusBadRequest, denied)
			return
		}

		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, gin.H{"error": err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.Page("tag_"+tag, entry))
	})

	router.GET("/api/tag/:tag/page/:entry/image", func(context *gin.Context) {
		tag := context.Param("tag")
		if !instance.MatchName(tag) {
			context.JSON(http.StatusBadRequest, denied)
			return
		}

		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, gin.H{"error": err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.PageImages("tag_"+tag, entry))
	})

	router.GET("/api/tag/:tag/info", func(context *gin.Context) {
		context.JSON(http.StatusOK, instance.TagInfo(context.Param("tag")))
	})

	router.PATCH("/api/tag/:tag/info", func(context *gin.Context) {
		info, ok := getUser(context)

		// Require sign in
		if !ok {
			context.JSON(http.StatusForbidden, unauthorized)
			return
		}

		if !info.Privileged {
			context.JSON(http.StatusForbidden, denied)
			return
		}

		var payload TagUpdatePayload
		if err := context.ShouldBindJSON(&payload); err != nil {
			context.JSON(http.StatusBadRequest, gin.H{
				"error": err.Error(),
			})
			return
		}
		instance.TagType(context.Param("tag"), payload.Type)
	})

	router.GET("/api/page", func(context *gin.Context) {
		context.String(http.StatusOK, strconv.Itoa(instance.PageTotal(store.ImageRootPageVariant)))
	})

	router.GET("/api/page/:entry", func(context *gin.Context) {
		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, gin.H{"error": err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.Page(store.ImageRootPageVariant, entry))
	})

	router.GET("/api/page/:entry/image", func(context *gin.Context) {
		param := context.Param("entry")
		entry, err := strconv.Atoi(param)
		if err != nil {
			context.JSON(http.StatusBadRequest, gin.H{"error": err.Error()})
			return
		}
		context.JSON(http.StatusOK, instance.PageImages(store.ImageRootPageVariant, entry))
	})
}

func getUser(context *gin.Context) (store.User, bool) {
	if instance.SingleUser {
		return instance.User(instance.InitialUser), true
	}
	secret := context.GetHeader("secret")
	info := instance.SecretLookup(secret)
	if info.Secret != secret || info.Snowflake == "" {
		return store.User{}, false
	}
	return info, true
}
